# -*- coding: utf-8 -*-
###########################################################################
# EOLE - 2013-2017
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
###########################################################################
"""Parsing des fichiers de données AAF
 - parse_aaf_eleves
 - parse_aaf_profs
 - parse_aaf_responsables
 - parse_aaf_etab_educnat
"""
from lxml import etree  # fast and furious xml library

from eoleaaf.config import aaf_type, ConfigError

from .config import dbtype

if dbtype in ["mysql", "sqlite"]:
    from eoleaaf.util import MinimalSQLGenerator as DataBaseGenerator
elif dbtype == "mongodb":
    from eoleaaf.util import MongoDBGenerator as DataBaseGenerator
    from eoleaaf.sambatranslator import gen_user_displayname


from eoleaaf.util import modif_fields, pwgen, db, \
                 parse_multi, get_new_id, gen_user_login
from eoleaaf.miscutil import user_id, classes_for_one_eleve, \
                                groupe_eleve
from logging import getLogger
aaflog = getLogger('AAF')

# compat python3
unicode = str

# mode (complet ou mise a jour) pour l'importation
# le mode d'utilisation peut être fixé de l'extérieur du module
maj_mode = False

# ____________________________________________________________
# buffers (pour rendre plus rapide les database queries)
# ATTENTION : ce sont des variables globales
classes_buffer = {}
groupes_buffer = {}
eleves_responsable_buffer = {}

# ____________________________________________________________
def is_empty(dico, cle):
    """
    pas de clé ou valeur vide
    """
    if dico.has_key(cle) and dico[cle] != '':
        return False
    else:
        return True


def not_empty(dico, cle):
    """
    verifie que la cle existe
    et que la valeur associee n'est pas vide
    """
    return not is_empty(dico, cle)


# ____________________________________________________________
DEBUG_NUM = 100

# ____________________________________________________________
# global (and unique) sql generator object
sqlgen = DataBaseGenerator()


#_______________________________________________________________________________
# sql inserts


def _get_oriname(uai, name):
    return u"{}_{}".format(uai, name)


def _get_groupname(uai, name):
    return _get_oriname(uai, name)


def _parse_etab_values(key, userdata, etab_buffer):
    def _parse_etab_value(value, etab_buffer):
        splitted = value.split('$')
        etab = splitted[0]
        name = splitted[-1]
        etablissement_id, uai = _mongo_fetch_value(etab, "etablissement", "ENTStructureJointure", etab_buffer)
        return {'etablissement': etablissement_id,
                'etablissement_name': uai,
                'name': _get_groupname(uai, name),
                'oriName': _get_oriname(uai, name)}
    if key in userdata:
        values = userdata[key]
        new_values = []
        if isinstance(values, list):
            new_values = []
            for value in values:
                if value != '':
                    new_values.append(_parse_etab_value(value, etab_buffer))
        elif values != '':
            new_values = _parse_etab_value(values, etab_buffer)
        userdata[key] = new_values


def _parse_functions(values, etab_buffer):
    def _parse_function(value, etab_buffer):
        splitted = value.split('$')
        etab = splitted[0]
        name = splitted[2]
        discipline = splitted[4]
        return {'etablissement': _mongo_fetch_value(etab, "etablissement", "ENTStructureJointure", etab_buffer)[0],
                'function': name,
                'discipline': discipline}
    if values == '':
        return values
    if isinstance(values, list):
        new_values = []
        for value in values:
            new_values.append(_parse_function(value, etab_buffer))
    else:
        new_values = _parse_function(values, etab_buffer)
    return new_values


def _parse_ens_class_subjects(values, etab_buffer, subject_buffer, name='class'):
    def _parse_ens_class_subject(value, etab_buffer, subject_buffer):
        splitted = value.split('$')
        etab = splitted[0]
        classe = splitted[1]
        subject = splitted[2]
        return {'etablissement': _mongo_fetch_value(etab, "etablissement", "ENTStructureJointure", etab_buffer)[0],
                name: classe,
                'subject': _mongo_fetch_value(subject, "subject", "ENTMatJointure", subject_buffer)}
    if values == '':
        return values
    if isinstance(values, list):
        new_values = []
        for value in values:
            new_values.append(_parse_ens_class_subject(value, etab_buffer, subject_buffer))
    else:
        new_values = _parse_ens_class_subject(values, etab_buffer, subject_buffer)
    return new_values


def _parse_attrib_split(value, idx):
    if value == '':
        new_value = value
    elif isinstance(value, list):
        new_value = []
        for val in value:
            new_value.append(val.split('$')[idx])
    else:
        new_value = value.split('$')[idx]
    return new_value


def _parse_aaf_profadmin(usertype, userdata):
    if usertype == 'administratif':
        unused_fields = ['ENTAuxEnsClassesPrincipal', 'ENTAuxEnsCategoDiscipline',
            'ENTAuxEnsDisciplinesPoste', 'ENTAuxEnsGroupes',
           'ENTAuxEnsGroupesMatieres']
        for field in unused_fields:
            if field in userdata:
                   del(userdata[field])
    if dbtype == "mongodb":
        etab_buffer = {}
        subject_buffer = {}
        _replace_id_by_reference(userdata, "ENTPersonStructRattach", "etablissement", "ENTStructureJointure", etab_buffer)
        if usertype == 'enseignant':
            for key in ['ENTAuxEnsMatiereEnseignEtab', 'ENTAuxEnsClasses', 'ENTAuxEnsMEF', 'ENTAuxEnsClassesPrincipal', 'ENTAuxEnsGroupes']:
                _parse_etab_values(key, userdata, etab_buffer)
            if 'ENTAuxEnsDisciplinesPoste' in userdata:
                userdata['ENTAuxEnsDisciplinesPoste'] = _parse_attrib_split(userdata['ENTAuxEnsDisciplinesPoste'], 0)
            if 'ENTAuxEnsCategoDiscipline' in userdata:
                userdata['ENTAuxEnsCategoDiscipline'] = _parse_attrib_split(userdata['ENTAuxEnsCategoDiscipline'], 1)
            if u"ENTAuxEnsClassesMatieres" in userdata:
                userdata[u"ENTAuxEnsClassesMatieres"] = _parse_ens_class_subjects(userdata[u"ENTAuxEnsClassesMatieres"], etab_buffer, subject_buffer)
            if u"ENTAuxEnsGroupesMatieres" in userdata:
                userdata[u"ENTAuxEnsGroupesMatieres"] = _parse_ens_class_subjects(userdata[u"ENTAuxEnsGroupesMatieres"], etab_buffer, subject_buffer, 'group')
        if 'ENTPersonFonctions' in userdata:
            userdata['ENTPersonFonctions'] = _parse_functions(userdata['ENTPersonFonctions'], etab_buffer)
        userdata['gecos'] = str(gen_user_displayname(userdata['sn'], userdata['givenName']))


def delete_profs(ids):
    db_profs = set()
    db_admins = set()
    for prof in db.fetchall('user', attr=u'UserType', value='enseignant', projection={'id':1}):
        db_profs.add(prof['id'])
    for admin in db.fetchall('user', attr=u'UserType', value='administratif', projection={'id':1}):
        db_admins.add(admin['id'])
    users_to_be_deleted = (db_profs | db_admins) - set(ids)
    for user in users_to_be_deleted:
        db_users = db.fetchone(None, '*', 'user', "id", user, sqlgen)
        if user in db_profs:
            usertype = 'enseignant'
        else:
            usertype = 'administratif'
        modif_fields(db_users, sqlgen, 'DELETE',
                     modif_type='DELTADELETE', tablename=usertype, fromxml=True, db=db,
                     no_bulk=True)
    db.commit('user', no_bulk=True)


def _mongo_fetch_value(value, tablename, filterby, tbuffer):
    if tbuffer is not None and value in tbuffer:
        new_value = tbuffer[value]
    else:
        fetch_result = db.fetchone(None, None, tablename, filterby, value, None)
        if fetch_result is not None:
            if tablename == 'etablissement':
                new_value = (fetch_result[u'_id'], fetch_result["ENTEtablissementUAI"])
            else:
                new_value = fetch_result[u'_id']
        else:
            if tablename == 'etablissement':
                new_value = (value, "-")
            else:
                new_value = value
        if tbuffer is not None:
            tbuffer[value] = new_value
    return new_value


def _mongo_fetch_values(values, tablename, filterby, tbuffer):
    if isinstance(values, list) and len(values) == 1:
        values = values[0]
    if values == '':
        return ''
    if isinstance(values, list):
        new_values = []
        for val in values:
            nvalues = _mongo_fetch_value(val, tablename, filterby, tbuffer)
            if tablename == 'etablissement':
                new_values.append(nvalues[0])
            else:
                new_values.append(nvalues)
    else:
        new_values = _mongo_fetch_value(values, tablename, filterby, tbuffer)
        if tablename == 'etablissement':
            new_values = new_values[0]
    return new_values


def _replace_id_by_reference(userdata, key, tablename, filterby, tbuffer=None):
    if key in userdata:
        value = _mongo_fetch_values(userdata[key], tablename, filterby, tbuffer)
        userdata[key] = value


def _parse_aaf_student(userdata, uid, classes_buffer, groupes_buffer, eleves_responsable_buffer):
    # ajout de l'eleve a la classe et au groupe
    if not dbtype == "mongodb":
        if aaf_type == "samba4":
            cn = userdata['ENTPersonLogin']
            classes_for_one_eleve(userdata, classes_buffer, cn)
            groupe_eleve(userdata, groupes_buffer, cn)
        elif aaf_type== "openldap":
            classes_for_one_eleve(userdata, classes_buffer, uid)
            groupe_eleve(userdata, groupes_buffer, uid)
        else:
            raise ConfigError("wrong or unknown aaf_type configuration value: {}".format(aaf_type))
    else:
        etab_buffer = {}
        if 'ENTEleveGroupes' in userdata:
            _parse_etab_values('ENTEleveGroupes', userdata, etab_buffer)
        resp_buffer = {}
        for key in ('ENTEleveParents', 'ENTElevePere', 'ENTEleveMere', 'ENTEleveAutoriteParentale'):
            _replace_id_by_reference(userdata, key, "user", "ENTPersonJointure", resp_buffer)
        _replace_id_by_reference(userdata, u'ENTEleveCodeEnseignements', "subject", "ENTMatJointure")

        etab_buffer = {}
        _replace_id_by_reference(userdata, "ENTPersonStructRattach", "etablissement", "ENTStructureJointure", etab_buffer)
        _parse_etab_values('ENTEleveClasses', userdata, etab_buffer)
        userdata['gecos'] = str(gen_user_displayname(userdata['sn'], userdata['givenName']))
    if isinstance(userdata['ENTElevePersRelEleve'], list):
        resps = userdata['ENTElevePersRelEleve']
    else:
        resps = userdata['ENTElevePersRelEleve'].split('\n')

    for resp in resps:
        numresp = resp.split('$')[0]
        if numresp in eleves_responsable_buffer:
            if uid not in eleves_responsable_buffer[numresp]:
                eleves_responsable_buffer[numresp].append(uid)
        else:
            eleves_responsable_buffer[numresp] = [uid]

def _parse_aaf_resp(userdata, eleves_responsable_buffer):
    # ____________________________________________________________
    # remplissage des élèves associé au userdata
    # à partir des inforamtions bufferisées
    if dbtype == 'mongodb':
        userdata['gecos'] = str(gen_user_displayname(userdata['sn'], userdata['givenName']))
    else:
        if userdata['id'] in eleves_responsable_buffer:
            userdata['ENTAuxPersRelEleveEleve'] = '\n'.join(eleves_responsable_buffer[userdata['id']])
        else:
            # FIXME: should not append and if it is the case
            # the account should not be created
            userdata['ENTAuxPersRelEleveEleve'] = None

def _modif_type_to_action(modif_type):
    if modif_type in ('CREATE', 'DELTACREATE'):
        action = 'ADD'
    elif modif_type in ('UPDATE', 'LDIFEXPORT', 'DELTAUPDATE', 'DELTAMOVE'):
        action = 'UPDATE'
    elif modif_type == 'DELTADELETE':
        action == 'DELETE'
    else:
        raise Exception('Unknown modif_type')
    return action

# ____________________________________________________________
# Generic parsing function
def parse_aaf(type_, filename, aaftype, modif_type, parse_type='ADD'):
    """parsing depuis AAF"""
    if modif_type == 'CREATE' and not maj_mode:
        #cas aaf-complet (non maj)
        no_bulk = False
    else:
        no_bulk = True
    num = 0
    aaflog.info("Lecture des {}s...".format(type_))
    login_buffer = []
    rnd_buffer = list()
    if parse_type == 'ADD':
        parse_tag = 'addRequest'
        parse_find = 'attributes/attr'

    elif parse_type == 'MODIFY':
        parse_tag = 'modifyRequest'
        parse_find = 'modifications/modification'

    action = _modif_type_to_action(modif_type)

    context = etree.iterparse(filename, events=('end',), tag=parse_tag)
    users_id = []
    for event, tdata in context:
        _type = tdata.find('operationalAttributes/attr/value').text
        if _type != aaftype:
            aaflog.info("%s pas un {} :(".format(_type, type_))
            continue
        #**Attention** identifier/id == ENTPersonJointure
        userdata = user_id(tdata)
        users_id.append(userdata['id'])
        for attr in tdata.findall(parse_find):
            balise, vals = parse_multi(attr)
            userdata[balise] = vals
        retrieved = False
        if type_ == 'administratif':
            #Règle de répartition des personnes
            #cf. Annexe 4 (Alimentation depuis le SI MEN) version 4.2
            if userdata['PersEducNatPresenceDevantEleves'] == 'O' or \
                   '$DOC$' in userdata['ENTPersonFonctions']:
                usertype = 'enseignant'
            else:
                usertype = 'administratif'
        else:
            usertype = type_
        if maj_mode or parse_type == 'MODIFY':
            db_user = db.fetchone(usertype, '*', "user", "id", userdata['id'], sqlgen)
            if db_user is None:
                modif_type = 'DELTACREATE'
                db_user = {}
                if parse_type == 'MODIFY':
                    aaflog.warn("ATTENTION: l'{1} {0} censé être mis a jour "
                             "n'existe pas dans la base".format(userdata['id'], type_))
                action = 'ADD'
            else:
                current_modif_type = db_user['FieldActionType']
                if current_modif_type in ['CREATE', 'DELTACREATE']:
                    modif_type = 'DELTACREATE'
                elif current_modif_type == 'DELTADELETE':
                    modif_type = 'DELTADELETE'
                retrieved = True
                action = 'MODIFY'
        else:
            db_user = {}
        if 'uid' not in db_user:
            uid = userdata['uid'] = get_new_id(rnd_buffer, sqlgen)
            if retrieved:
                aaflog.error("ATTENTION: l'eleve {0} aurait dû avoir un "
                          "uid".format(userdata['id']))
        else:
            uid = userdata['uid'] = db_user['uid']

        if 'userPassword' not in db_user:
            userdata['userPassword'] = pwgen()
            if retrieved:
                aaflog.error("ATTENTION: l'eleve {0} aurait dû avoir un mot de "
                          "passe".format(userdata['id']))
        else:
            userdata['userPassword'] = db_user['userPassword']

        # **IMPORTANT** : this login shall not change once set at the user's creation
        # from the AAF point of vue, the key is the id
        # for samba, it's the generated cn
        if 'ENTPersonLogin' not in db_user:
            if retrieved:
                aaflog.error("ATTENTION: l'eleve {0} aurait dû avoir un "
                          "login".format(userdata['id']))
            userdata['ENTPersonLogin'] = gen_user_login(userdata['sn'],
                                                        userdata['givenName'], uid, type_,
                                                        userdata['id'], sqlgen, login_buffer)
        else:
            userdata['ENTPersonLogin'] = db_user['ENTPersonLogin']

        if usertype == 'eleve':
            # TODO: on avait évoqué la possibilité de pouvoir limiter la création
            # aux responsables "1" & "2" (resp.split('$')[1] in ['1', '2'])
            # ce qui revient à être dans la liste
            # de "ENTEleveAutoriteParentale" (cf. Scribe)
            if 'ENTElevePersRelEleve' not in userdata:
                #log.error("ENTElevePersRelEleve not present in {0}".format(eleve['id']))
                continue
            _parse_aaf_student(userdata, uid, classes_buffer, groupes_buffer, eleves_responsable_buffer)
        elif type_ == 'responsable':
            _parse_aaf_resp(userdata, eleves_responsable_buffer)
        elif type_ == 'administratif':
            _parse_aaf_profadmin(usertype, userdata)

        launch_modif_fields = True
        if retrieved:
            unlisted_key = set([u'FieldFromXMLImportActionDate', u'_id', u'FieldModifiedStatus', u'FieldActionType', u'UserType'])
            key_old = set(db_user.keys())
            key_new = set(userdata.keys())
            is_modified = False
            if len(key_old - key_new - unlisted_key) != 0 or len(key_new - key_old) != 0:
                is_modified = True
                aaflog.debug("some fields in the database ({}) and the current user ({}) does not match for the user: {}".format(
                          key_old - unlisted_key, key_new, userdata['id']))
            else:
                for key in key_new:
                    if db_user[key] != userdata[key]:
                        is_modified = True
                        aaflog.debug("some entries in the database ({}) and the current user ({}) does not match for the user".format(
                                  db_user[key], userdata[key], userdata['id']))
                        break
            if is_modified:
                if db_user['FieldActionType'] == 'DELTADELETE':
                    pass
                elif db_user['FieldActionType'] in ['CREATE', 'DELTACREATE']:
                    modif_type = 'DELTACREATE'
                else:
                    if db_user.get('ENTPersonStructRattach') != userdata.get('ENTPersonStructRattach'):
                        modif_type = 'DELTAMOVE'
                        if 'ENTPersonStructRattachsOld' not in userdata:
                            userdata['ENTPersonStructRattachsOld'] = []
                        userdata['ENTPersonStructRattachsOld'].append(db_user.get('ENTPersonStructRattach'))
                    else:
                        modif_type = 'DELTAUPDATE'
                action = 'UPDATE'
            else:
                launch_modif_fields = False
        if launch_modif_fields:
            # ____________________________________________________________
            # traitement de l'utilisateur
            modif_fields(userdata, sqlgen, action, modif_type, tablename=usertype,
                        fromxml=True, db=db, no_bulk=no_bulk)
        num += 1
        if num % DEBUG_NUM == 0:
            aaflog.debug("{} {}s lus...".format(num, type_))
        tdata.clear()
        while tdata.getprevious() is not None:
            del tdata.getparent()[0]
    del context
    login_buffer = []

    db.commit("user", no_bulk=no_bulk)
    aaflog.info("TOTAL : {} {}s".format(num, type_))
    return users_id

# ____________________________________________________________
# Extraction AAF Eleves et Responsables
def parse_aaf_eleves(eleve_file, modif_type='CREATE', parse_type='ADD'):
    """parsing des élèves depuis AAF"""
    return parse_aaf('eleve', eleve_file, 'Eleve', modif_type, parse_type)


def delete_eleves(ids):
    db_users = set()
    for user in db.fetchall('user', attr=u'UserType', value='eleve', projection={'id':1}):
        db_users.add(user['id'])
    users_to_be_deleted = db_users - set(ids)
    for user in users_to_be_deleted:
        db_user = db.fetchone(None, '*', 'user', "id", user, sqlgen)
        modif_fields(db_user, sqlgen, 'DELETE', modif_type='DELTADELETE', tablename='eleve',
                     fromxml=True, db=db, no_bulk=True)
    db.commit('user', no_bulk=True)


def parse_aaf_responsables(responsables_file, modif_type='CREATE', parse_type='ADD'):
    """Parsing XML des responsables depuis AAF
    :responsables_file: file name like ``*_PersRelEleve_*.xml``
    """
    return parse_aaf('responsable', responsables_file, 'PersRelEleve', modif_type, parse_type)


def delete_responsables(ids):
    db_users = set()
    for user in db.fetchall('user', attr=u'UserType', value='responsable', projection={'id':1}):
        db_users.add(user['id'])
    users_to_be_deleted = db_users - set(ids)
    for user in users_to_be_deleted:
        db_user = db.fetchone(None, '*', 'user', "id", user, sqlgen)
        modif_fields(db_user, sqlgen, 'DELETE', modif_type='DELTADELETE', tablename='responsable',
                     fromxml=True, db=db, no_bulk=True)
    db.commit('user', no_bulk=True)


def parse_aaf_profs(profs_file, modif_type='CREATE', parse_type='ADD'):
    """ parsing des professeurs depuis AAF
    Extraction AAF Professeurs + Administratifs
    Fichier de type : ``ENT_$RNE_Complet_$DATE_PersEducNat_0000.xml``
    """
    return parse_aaf('administratif', profs_file, 'PersEducNat', modif_type, parse_type)


# ____________________________________________________________
def parse_aaf_etab_educnat(etab_file, modif_type='CREATE', parse_type='ADD'):
    """Parsing XML des etablissements depuis AAF
    :etab_file: file name like ``*_EtabEducNat_*.xml``
    """
    # init
    if modif_type == 'CREATE' and not maj_mode:
        #cas aaf-complet (non maj)
        no_bulk = False
    else:
        no_bulk = True
    num = 0
    action = _modif_type_to_action(modif_type)
    aaflog.info("Lecture des etablissements...")
    etablissements_id = []
    if parse_type == 'ADD':
        parse_tag = 'addRequest'
        parse_find = 'attributes/attr'

    elif parse_type == 'MODIFY':
        parse_tag = 'modifyRequest'
        parse_find = 'modifications/modification'

    context = etree.iterparse(etab_file, events=('end',), tag=parse_tag)
    # ____________________________________________________________
    # dict construction by XML parsing
    for event, tresp in context:
        _type = tresp.find('operationalAttributes/attr/value').text
        if _type != 'EtabEducNat':
            aaflog.info("%s n'est pas un etablissement :(" % _type)
            continue
        etablissement = dict()
        etabid = tresp.find('identifier/id').text
        etablissement['id'] = unicode(etabid)
        etablissements_id.append(etablissement['id'])
        for attr in tresp.findall(parse_find):
            balise, vals = parse_multi(attr)
            etablissement[balise] = vals
        retrieved = False
        if maj_mode or parse_type == 'MODIFY':
            db_etab = db.fetchone(None, '*', 'etablissement', "id", etablissement['id'], sqlgen)
            if db_etab is None:
                modif_type = 'DELTACREATE'
                if not maj_mode:
                    aaflog.warn("ATTENTION: l'établissement {0} censé être mis a jour "
                             "n'existe pas dans la base".format(etablissement['id']))
                action = 'ADD'
            else:
                current_modif_type = db_etab['FieldActionType']
                if maj_mode:
                    modif_type = 'DELTAUPDATE'
                else:
                    if current_modif_type in ['CREATE', 'DELTACREATE']:
                        modif_type = 'DELTACREATE'
                    elif current_modif_type == 'DELTADELETE':
                        modif_type = 'DELTADELETE'
                retrieved = True
                action = 'UPDATE'
        if etablissement['ENTStructureSIREN'] == '':
            aaflog.error("Pas de SIREN pour l'établissement {0}".format(etablissement['ENTStructureUAI']))
        if isinstance(etablissement['ENTStructureGroupes'], list):
            new_value = []
            for value in etablissement['ENTStructureGroupes']:
                spl = value.split('$')
                new_value.append({'class': spl[-1], 'group': spl[0]})
            etablissement['ENTStructureGroupes'] = new_value
        if u"ENTEtablissementBassin" in etablissement and '$' in etablissement[u"ENTEtablissementBassin"]:
            etablissement[u"ENTEtablissementBassin"] = etablissement[u"ENTEtablissementBassin"].split('$')[1]

        launch_modif_fields = True
        if retrieved:
            unlisted_key = set([u'FieldFromXMLImportActionDate', u'_id', u'FieldModifiedStatus', u'FieldActionType'])
            key_old = set(db_etab.keys())
            key_new = set(etablissement.keys())
            is_modified = False
            if len(key_old - key_new - unlisted_key) != 0 or len(key_new - key_old) != 0:
                is_modified = True
                aaflog.debug("some fields in the database ({}) and the current etablissement ({}) does not match for the etablissement: {}".format(
                          key_old - unlisted_key, key_new, etablissement['ENTStructureUAI']))
            else:
                for key in key_new:
                    if db_etab[key] != etablissement[key]:
                        is_modified = True
                        aaflog.debug("some entries in the database ({}) and the current etablissement ({}) does not match for the etablissement".format(
                                  db_etab[key], etablissement[key], etablissement['ENTStructureUAI']))
                        break
            if is_modified:
                if db_etab[u'FieldActionType'] in ['CREATE', 'DELTACREATE']:
                    modif_type = 'DELTACREATE'
                else:
                    modif_type = 'DELTAUPDATE'
                action = 'UPDATE'
            else:
                launch_modif_fields = False
        if launch_modif_fields:
            # ____________________________________________________________
            # database insert
            modif_fields(etablissement, sqlgen, action, modif_type=modif_type, tablename="etablissement",
                        fromxml=True, db=db, no_bulk=no_bulk)
        # ____________________________________________________________
        num += 1
        if num % DEBUG_NUM == 0:
            aaflog.debug("%d etablissements lus..." % num)

    db.commit('etablissement', no_bulk=no_bulk)
    aaflog.info("TOTAL : %d etablissements" % num)
    return etablissements_id

def delete_etab_educnat(ids):
    db_etabs = set()
    for etab in db.fetchall('etablissement', projection={'id':1}):
        db_etabs.add(etab['id'])
    etab_to_be_deleted = db_etabs - set(ids)
    for etab in etab_to_be_deleted:
        db_etab = db.fetchone(None, '*', 'etablissement', "id", etab, sqlgen)
        modif_fields(db_etab, sqlgen, 'DELETE', modif_type='DELTADELETE', tablename="etablissement",
                     fromxml=True, db=db, no_bulk=True)
    db.commit('etablissement', no_bulk=True)


def parse_aaf_subjects(subject_file, modif_type='CREATE', parse_type='ADD'):
    """Parsing XML des matières depuis AAF
    :etab_file: file name like ``*_MatiereEducNat_*.xml``
    """
    # init
    if modif_type == 'CREATE' and not maj_mode:
        #cas aaf-complet (non maj)
        no_bulk = False
    else:
        no_bulk = True
    num = 0
    aaflog.info("Lecture des matieres...")
    subject_ids = []
    if parse_type == 'ADD':
        parse_tag = 'addRequest'
        parse_find = 'attributes/attr'

    elif parse_type == 'MODIFY':
        parse_tag = 'modifyRequest'
        parse_find = 'modifications/modification'

    action = _modif_type_to_action(modif_type)

    context = etree.iterparse(subject_file, events=('end',), tag=parse_tag)
    # ____________________________________________________________
    # dict construction by XML parsing
    for event, tresp in context:
        subject = dict()
        _type = tresp.find('operationalAttributes/attr/value').text
        if _type != 'MatEducNat':
            aaflog.info("%s n'est pas une matiere :(" % _type)
            continue
        subjectid = tresp.find('identifier/id').text
        subject['id'] = unicode(subjectid)
        subject_ids.append(subject['id'])
        for attr in tresp.findall(parse_find):
            balise, vals = parse_multi(attr)
            subject[balise] = vals
        retrieved = False
        launch_modif_fields = True
        if maj_mode or parse_type == 'MODIFY':
            db_subject = db.fetchone(None, '*', 'subject', "id", subject['id'], sqlgen)
            if db_subject is None:
                modif_type = 'DELTACREATE'
                if parse_type == 'MODIFY':
                    aaflog.warn("ATTENTION: la matière {0} censée être mise a jour "
                             "n'existe pas dans la base".format(subject['id']))
                action = 'ADD'
            else:
                current_modif_type = db_subject['FieldActionType']
                if current_modif_type in ['CREATE', 'DELTACREATE']:
                    modif_type = 'DELTACREATE'
                elif current_modif_type == 'DELTADELETE':
                    modif_type = 'DELTADELETE'
                retrieved = True
                action = 'UPDATE'
        if retrieved:
            unlisted_key = set([u'FieldFromXMLImportActionDate', u'_id', u'FieldModifiedStatus', u'FieldActionType'])
            key_old = set(db_subject.keys())
            key_new = set(subject.keys())
            is_modified = False
            if len(key_old - key_new - unlisted_key) != 0 or len(key_new - key_old) != 0:
                is_modified = True
                aaflog.debug("some fields in the database ({}) and the current subject ({}) does not match for the subject: {}".format(
                          key_old - unlisted_key, key_new, subject['id']))
            else:
                for key in key_new:
                    if db_subject[key] != subject[key]:
                        is_modified = True
                        aaflog.debug("some entries in the database ({}) and the current subject ({}) does not match for the subject".format(
                                  db_subject[key], subject[key], subject['id']))
                        break
            if is_modified:
                if db_subject[u'FieldActionType'] in ['CREATE', 'DELTACREATE']:
                    modif_type = 'DELTACREATE'
                else:
                    modif_type = 'DELTAUPDATE'
                action = 'UPDATE'
            else:
                launch_modif_fields = False
        if launch_modif_fields:
            # database insert
            modif_fields(subject, sqlgen, action, modif_type=modif_type, tablename="subject",
                        fromxml=True, db=db, no_bulk=no_bulk)
        ## ____________________________________________________________
        num += 1
        if num % DEBUG_NUM == 0:
            aaflog.debug("%d matières lues..." % num)

    db.commit('subject', no_bulk=no_bulk)
    aaflog.info("TOTAL : %d matières" % num)
    return subject_ids

def delete_subject(ids):
    db_subjects = set()
    for subject in db.fetchall('subject', projection={'id':1}):
        db_subjects.add(subject['id'])
    subject_to_be_deleted = db_subjects - set(ids)
    for subject in subject_to_be_deleted:
        db_subject = db.fetchone(None, '*', 'subject', "id", subject, sqlgen)
        modif_fields(db_subject, sqlgen, 'DELETE', modif_type='DELTADELETE', tablename="subject",
                     fromxml=True, db=db, no_bulk=True)
    db.commit('subject', no_bulk=True)


# ____________________________________________________________
def reset_db():
    aaflog.info('Suppression de toutes les entrées de la base de données')
    db.reset()
    db.commit(None)
